﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using HeuristicLab.Collections;
using HeuristicLab.Core;
using HeuristicLab.Encodings.PackingEncoding.GroupingVector;
using HeuristicLab.Encodings.PackingEncoding.MultiComponentVector;
using HeuristicLab.Encodings.PackingEncoding.PackingPlan;
using HeuristicLab.Encodings.PackingEncoding.PackingSequence;
using HeuristicLab.Encodings.PermutationEncoding;
using HeuristicLab.Problems.BinPacking.Dimensions;
using HeuristicLab.Problems.BinPacking.Interfaces;
using HeuristicLab.Problems.BinPacking.PackingBin;
using HeuristicLab.Problems.BinPacking.PackingItem;
using HeuristicLab.Problems.BinPacking.Shapes;

namespace HeuristicLab.Problems.BinPacking.Decoders {
  public static class TwoDimensionalPackingFunctions {
    
    public static ObservableList<BinPacking<TwoDimensionalPacking, RectangularPackingBin, RectangularPackingItem>> BottomLeftPacking(
      MultiComponentVectorEncoding solution, RectangularPackingBin binMeasures, ItemList<RectangularPackingItem> itemMeasures) {
      
      ObservableList<BinPacking<TwoDimensionalPacking, RectangularPackingBin, RectangularPackingItem>> result = 
        new ObservableList<BinPacking<TwoDimensionalPacking, RectangularPackingBin, RectangularPackingItem>>();

      var sequenceMatrix = SequentialPackingFunctions.GenerateSequenceMatrix(solution);
      Dictionary<int, bool> rotated = SequentialPackingFunctions.GenerateRotationArray(solution);
          
      //Fill bins according to grouping vector
      List<int> remainingIDs = new List<int>();
      foreach (var sequence in sequenceMatrix) {
        remainingIDs = remainingIDs.Concat(sequence).ToList();
        result.Add(SequentialPackingFunctions.BottomLeftPacking(ref remainingIDs, binMeasures, itemMeasures, rotated));
      }

      //Try to put remaining items in existing bins
      var temp = new List<int>(remainingIDs);
      foreach (int id in temp) {
        foreach (var bp in result) {  
          var position = bp.FindPositionBySliding(itemMeasures[id], rotated[id]);
          if (position != null) {
            bp.PackItem(id, itemMeasures[id], position);
            remainingIDs.Remove(id);
          }
        }
      }

      //Put still remaining items in new bins
      while (remainingIDs.Count > 0) {
        result.Add(SequentialPackingFunctions.BottomLeftPacking(ref remainingIDs, binMeasures, itemMeasures));
      }

      return result;
    }         

    public static ObservableList<BinPacking<TwoDimensionalPacking, RectangularPackingBin, RectangularPackingItem>> BottomLeftPacking(
      GroupingVectorEncoding solution, RectangularPackingBin binMeasures, ItemList<RectangularPackingItem> itemMeasures) {
      
      ObservableList<BinPacking<TwoDimensionalPacking, RectangularPackingBin, RectangularPackingItem>> result = 
        new ObservableList<BinPacking<TwoDimensionalPacking, RectangularPackingBin, RectangularPackingItem>>();

      var sequenceMatrix = SequentialPackingFunctions.GenerateSequenceMatrix(solution);
          
      //Fill bins according to grouping vector
      List<int> remainingIDs = new List<int>();
      foreach (var sequence in sequenceMatrix) {
        remainingIDs = remainingIDs.Concat(sequence).ToList();
        result.Add(SequentialPackingFunctions.BottomLeftPacking(ref remainingIDs, binMeasures, itemMeasures));
      }

      //Try to put remaining items in existing bins
      var temp = new List<int>(remainingIDs);
      foreach (int id in temp) {
        foreach (var bp in result) {
          var position = bp.FindPositionBySliding(itemMeasures[id], false);
          if (position != null) {
            bp.PackItem(id, itemMeasures[id], position);
            remainingIDs.Remove(id);
            break;
          }
        }
      }

      //Put still remaining items in new bins
      while (remainingIDs.Count > 0) {
        result.Add(SequentialPackingFunctions.BottomLeftPacking(ref remainingIDs, binMeasures, itemMeasures));
      }

      return result;
    }

    public static ObservableList<BinPacking<TwoDimensionalPacking, RectangularPackingBin, RectangularPackingItem>> BottomLeftPacking(
      PackingSequenceEncoding solution, RectangularPackingBin binMeasures, ItemList<RectangularPackingItem> itemMeasures) {

      ObservableList<BinPacking<TwoDimensionalPacking, RectangularPackingBin, RectangularPackingItem>> result = 
        new ObservableList<BinPacking<TwoDimensionalPacking, RectangularPackingBin, RectangularPackingItem>>();

      List<int> remainingIDs = new List<int>(solution.PackingSequence);
      while (remainingIDs.Count > 0) {
        result.Add(SequentialPackingFunctions.BottomLeftPacking(ref remainingIDs, binMeasures, itemMeasures));
      }

      return result;
    }



    public static ObservableList<BinPacking<TwoDimensionalPacking, RectangularPackingBin, RectangularPackingItem>> ExtremePointBasedPacking(MultiComponentVectorEncoding solution,
        ItemList<RectangularPackingItem> itemMeasures, RectangularPackingBin binMeasures) {

      ObservableList<BinPacking<TwoDimensionalPacking, RectangularPackingBin, RectangularPackingItem>> result =
        new ObservableList<BinPacking<TwoDimensionalPacking, RectangularPackingBin, RectangularPackingItem>>();

      var sequenceMatrix = SequentialPackingFunctions.GenerateSequenceMatrix(solution);
      Dictionary<int, bool> rotated = SequentialPackingFunctions.GenerateRotationArray(solution);

      //Fill bins according to grouping vector
      List<int> remainingIDs = new List<int>();
      foreach (var sequence in sequenceMatrix) {
        remainingIDs = remainingIDs.Concat(sequence).ToList();
        result.Add(SequentialPackingFunctions.ExtremePointBasedPacking(ref remainingIDs, binMeasures, itemMeasures, rotated));
      }

      //Try to put remaining items in existing bins
      var temp = new List<int>(remainingIDs);
      foreach (int id in temp) {
        foreach (var bp in result) {
          var position = bp.FindExtremePointForItem(itemMeasures[id], rotated[id], false);
          if (position != null) {
            bp.PackItem(id, itemMeasures[id], position);
            remainingIDs.Remove(id);
          }
        }
      }

      //Put still remaining items in new bins
      while (remainingIDs.Count > 0) {
        result.Add(SequentialPackingFunctions.ExtremePointBasedPacking(ref remainingIDs, binMeasures, itemMeasures));
      }

      return result;
    }

    public static ObservableList<BinPacking<TwoDimensionalPacking, RectangularPackingBin, RectangularPackingItem>> ExtremePointBasedPacking(GroupingVectorEncoding solution,
        ItemList<RectangularPackingItem> itemMeasures, RectangularPackingBin binMeasures) {

      ObservableList<BinPacking<TwoDimensionalPacking, RectangularPackingBin, RectangularPackingItem>> result =
        new ObservableList<BinPacking<TwoDimensionalPacking, RectangularPackingBin, RectangularPackingItem>>();

      var sequenceMatrix = SequentialPackingFunctions.GenerateSequenceMatrix(solution);

      //Fill bins according to grouping vector
      List<int> remainingIDs = new List<int>();
      foreach (var sequence in sequenceMatrix) {
        remainingIDs = remainingIDs.Concat(sequence).ToList();
        result.Add(SequentialPackingFunctions.ExtremePointBasedPacking(ref remainingIDs, binMeasures, itemMeasures));
      }

      //Try to put remaining items in existing bins
      var temp = new List<int>(remainingIDs);
      foreach (int id in temp) {
        foreach (var bp in result) {
          var position = bp.FindExtremePointForItem(itemMeasures[id], false, false);
          if (position != null) {
            bp.PackItem(id, itemMeasures[id], position);
            remainingIDs.Remove(id);
            break;
          }
        }
      }

      //Put still remaining items in new bins
      while (remainingIDs.Count > 0) {
        result.Add(SequentialPackingFunctions.ExtremePointBasedPacking(ref remainingIDs, binMeasures, itemMeasures));
      }

      return result;
    }         

    public static ObservableList<BinPacking<TwoDimensionalPacking, RectangularPackingBin, RectangularPackingItem>> ExtremePointBasedPacking(PackingSequenceEncoding solution,
        ItemList<RectangularPackingItem> itemMeasures, RectangularPackingBin binMeasures) {  

     ObservableList<BinPacking<TwoDimensionalPacking, RectangularPackingBin, RectangularPackingItem>> result = 
       new ObservableList<BinPacking<TwoDimensionalPacking, RectangularPackingBin, RectangularPackingItem>>();
                                 

     List<int> remainingIDs = new List<int>(solution.PackingSequence);
     while (remainingIDs.Count > 0) {
       result.Add(SequentialPackingFunctions.ExtremePointBasedPacking(ref remainingIDs, binMeasures, itemMeasures));
     }    
 
     return result;
   }
  }
}
